#!/usr/bin/ruby -w
# -*- ruby -*-

#     Attribute codes:
#         00=none 01=bold 04=underscore 05=blink 07=reverse 08=concealed
#     Text color codes:
#         30=black 31=red 32=green 33=yellow 34=blue 35=magenta 36=cyan 37=white
#     Background color codes:
#         40=black 41=red 42=green 43=yellow 44=blue 45=magenta 46=cyan 47=white

module ANSIColor

  # Adds the escape sequence for value retrieval.
  class ANSIEscapedColors < Hash
    def [](str)
      val = super
      if val
        val = "\e[" + val + "m"
      end
      val
    end
  end

  ATTRIBUTE_COLORS = ANSIEscapedColors[
    'none'       => '00', 
    'bold'       => '01',
    'underscore' => '04',
    'underline'  => '04',
    'blink'      => '05',
    'reverse'    => '07',
    'concealed'  => '08'
  ]

  TEXT_COLORS = ANSIEscapedColors[
    'black'   => '30',
    'red'     => '31',
    'green'   => '32',
    'yellow'  => '33',
    'blue'    => '34',
    'magenta' => '35',
    'cyan'    => '36',
    'white'   => '37'
  ]
  
  BACKGROUND_COLORS = ANSIEscapedColors[
    'black'   => '40',
    'red'     => '41',
    'green'   => '42',
    'yellow'  => '43',
    'blue'    => '44',
    'magenta' => '45',
    'cyan'    => '46',
    'white'   => '47'
  ]

  RESET = "\e[0m"

  FOREGROUND_COLORS = ATTRIBUTE_COLORS; FOREGROUND_COLORS.update(TEXT_COLORS)

  # returns the code for the given color string, which is in the format:
  # [foreground] on [background]. Note that the foreground and background sections
  # can have modifiers (attributes). Examples:
  #     black
  #     blue on white
  #     bold green on yellow
  #     underscore bold magenta on cyan
  #     underscore red on bold cyan

  def ANSIColor.code(str)
    fg, bg = str.split(/\s*\bon_?\s*/)
    foreground(fg) + background(bg)
  end

  # returns the code for the given background color(s)
  def ANSIColor.background(bgcolor)
    make_code(bgcolor, BACKGROUND_COLORS)
  end

  # returns the code for the given foreground color(s)
  def ANSIColor.foreground(fgcolor)
    make_code(fgcolor, FOREGROUND_COLORS)
  end

  def ANSIColor.reset
    RESET
  end

  protected

  def ANSIColor.make_code(str, colors)
    code = ""
    (str || "").split.each do |s|
      if colors[s]
        code += colors[s]
      else
        $stderr.puts "WARNING: ANSIColor::make_code(" + str + "): unknown color: " + s
        break
      end
    end
    code
  end

end


# Very minimal logging output. If verbose is set, this displays the method and
# line number whence called. It can be a mixin to a class, which displays the
# class and method from where it called. If not in a class, it displays only the
# method.

# All kids love log.

module Log
  
  @@width = 0
  @@output = $stdout
  @@fmt = "[%s:%04d] {%s}"
  @@autoalign = false
  @@colors = []
  @@colorize_line = false
  
  DEBUG = 0
  INFO  = 1
  WARN  = 2
  ERROR = 3
  FATAL = 4

  @@level = FATAL

  def Log.verbose=(v)
    @@level = v ? DEBUG : FATAL
  end

  def Log.level=(lvl)
    @@level = lvl
  end

  # Assigns output to the given stream.
  def Log.output=(io)
    @@output = io
  end

  # sets whether to colorize the entire line, or just the message.
  def Log.colorize_line=(col)
    @@colorize_line = col
  end

  # Assigns output to a file with the given name. Returns the file; client
  # is responsible for closing it.
  def Log.outfile=(fname)
    file = File.new(fname, "w")
    Log.output = file
    return file
  end

  # Creates a printf format for the given widths, for aligning output.
  def Log.set_widths(file_width, line_width, func_width)
    @@fmt = "[%#{file_width}s:%#{line_width}d] {%#{func_width}s}"
  end

  # Sets auto-align of the {function} section.  set_widths will likely
  # result in nicer output.
  def Log.autoalign
    @@autoalign = true
  end

  # Logs the given message, including the class whence invoked.
  def log(msg = "", level = DEBUG, depth = 1, &blk)
    Log.log(msg, level, depth + 1, self.class.to_s)
  end

  def debug(msg = "", depth = 1, &blk)
    Log.log(msg, DEBUG, depth + 1, self.class.to_s, &blk)
  end

  def info(msg = "", depth = 1, &blk)
    Log.log(msg, INFO, depth + 1, self.class.to_s, &blk)
  end

  def warn(msg = "", depth = 1, &blk)
    Log.log(msg, WARN, depth + 1, self.class.to_s, &blk)
  end

  def error(msg = "", depth = 1, &blk)
    Log.log(msg, ERROR, depth + 1, self.class.to_s, &blk)
  end

  def fatal(msg = "", depth = 1, &blk)
    Log.log(msg, FATAL, depth + 1, self.class.to_s, &blk)
  end

  def Log.debug(msg = "", depth = 1, &blk)
    Log.log(msg, DEBUG, depth + 1, &blk)
  end

  def Log.info(msg = "", depth = 1, &blk)
    Log.log(msg, INFO, depth + 1, &blk)
  end

  def Log.warn(msg = "", depth = 1, &blk)
    Log.log(msg, WARN, depth + 1, &blk)
  end

  def Log.error(msg = "", depth = 1, &blk)
    Log.log(msg, ERROR, depth + 1, &blk)
  end

  def Log.fatal(msg = "", depth = 1, &blk)
    Log.log(msg, FATAL, depth + 1, &blk)
  end

  # Logs the given message.
  def Log.log(msg = "", level = DEBUG, depth = 1, cname = nil, &blk)
    if level >= @@level
      c = caller(depth)[0]
      c.index(/(.*):(\d+)(?::in \`(.*)\')?/)
      file, line, func = $1, $2, $3
      if cname
        func = cname + "#" + func
      end
      file.sub!(/.*\//, "")
      if @@autoalign
        Log.print_autoaligned(file, line, func, msg, level, &blk)
      else
        Log.print_formatted(file, line, func, msg, level, &blk)
      end
    end
  end

  def Log.print_autoaligned(file, line, func, msg, level, &blk)
    @@width = [ @@width, func.length ].max
    hdr = sprintf "[%s:%04d] {%-*s} ", file, line, @@width
    Log.print(hdr, msg, level)
  end

  def Log.print_formatted(file, line, func, msg, level, &blk)
    hdr = sprintf @@fmt, file, line, func
    Log.print(hdr, msg, level, &blk)
  end

  
  def Log.print(hdr, msg, level, &blk)
    # puts "hdr: #{hdr}, msg #{msg}, level #{level}"
    if blk
      # puts "block is given"
      x = blk.call
      if x.kind_of?(String)
        # puts "got a string!"
        msg = x
      else
        # puts "did not get a string"
        return
      end
    else
      # puts "block not given"
    end

    if @@colors[level]
      if @@colorize_line
        @@output.puts @@colors[level] + hdr + " " + msg.to_s.chomp + ANSIColor.reset
      else
        @@output.puts hdr + " " + @@colors[level] + msg.to_s.chomp + ANSIColor.reset
      end
    else
      @@output.puts hdr + " " + msg.to_s.chomp
    end      
  end

  def Log.set_color(level, color)
    # log("#{level}, #{color}")
    @@colors[level] = ANSIColor::code(color)
  end

end

if __FILE__ == $0
  Log.verbose = true
  Log.set_widths 15, -5, -35
  #Log.outfile = "/tmp/log." + $$.to_s

  class Demo
    include Log
    
    def initialize
      # log "hello"
      Log.set_color(Log::DEBUG, "cyan")
      Log.set_color(Log::INFO,  "bold cyan") 
      Log.set_color(Log::WARN,  "reverse")
      Log.set_color(Log::ERROR, "bold red")
      Log.set_color(Log::FATAL, "bold white on red")
    end

    def meth
      # log

      i = 4
      # info { "i: #{i}" }

      i /= 3
      debug { "i: #{i}" }

      i **= 3
      info "i: #{i}"

      i **= 2
      warn "i: #{i}"

      i <<= 4
      error "i: #{i}"

      i <<= 1
      fatal "i: #{i}"
    end

  end

  class Another
    include Log
    
    def Another.cmeth
      # /// "Log" only in instance methods
      # log "I'm sorry, Dave, I'm afraid I can't do that."

      # But this is legal.
      Log.log "happy, happy, joy, joy"
    end
  end

  demo = Demo.new
  demo.meth

  # Log.colorize_line = true

  # demo.meth
  # Another.cmeth

  # Log.info "we are done."
  
end
